/*
 * Copyright (C) 2012 Oron Peled
 * Derived from su for GNU.
 * Copyright (C) 1992-2006, 2008 Free Software Foundation, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * This work was derived from proxy_pam:
 *    http://redmine.lighttpd.net/attachments/919/proxy_pam.c
 */

#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <xtalk/debug.h>
#include "xpanel-auth.h"

#if	HAVE_PAM
#include <security/pam_appl.h>
#include <security/pam_misc.h>
static int do_authenticate(const char *user, const char *pass);
#endif


int xpanel_auth_have_pam(void)
{
#if	HAVE_PAM
	return 1;
#else
	return 0;
#endif
}

/*
 * Return value: true/false
 */
int xpanel_auth_skip(const char *user_name)
{
	struct passwd *passwd;

	/* INFO("%s: user_name=%s\n", __func__, user_name); */
	passwd = getpwnam(user_name);
	if (!passwd)
		return 1;
	return 0;
}

/*
 * Return value:
 *      0  : OK
 *      <0 : -errno
 */
int xpanel_auth_check(const char *user_name, const char *pw_str)
{
#if	HAVE_PAM
	int ret;

	ret = do_authenticate(user_name, pw_str);
	if(ret != 0) {
		ERR("Authentication for %s failed: %d\n", user_name, ret);
		return -EPERM;
	}
	return 0;
#else
	return (strcmp(pw_str, "42") == 0) ? 0 : -EPERM;
#endif
}

#if	HAVE_PAM

int fake_conv(int num_msg, const struct pam_message **msg, struct pam_response **resp, void *appdata_ptr)
{
	char			*pass = appdata_ptr;
	struct pam_response	*response = calloc(1, sizeof(*response));

	
	response->resp = pass;
	response->resp_retcode = 0;
	*resp = response;
	//ERR("euid=%d\n", geteuid());
	//ERR("num_msg=%d msg[0]='%s' pass='%s'\n", num_msg, (*msg)->msg, pass);
	return PAM_SUCCESS;
}

static struct pam_conv conv = {
  fake_conv,
  NULL
};

/* Ask the user for a password.
   If PAM is in use, let PAM ask for the password if necessary.
   Return true if the user gives the correct password for entry PW,
   false if not.  Return true without asking for a password if run by UID 0
   or if PW has an empty password.  */

static int do_authenticate(const char *user, const char *pass)
{
	pam_handle_t	*pamh = NULL;
	int		retval;

	conv.appdata_ptr = strdup(pass);
	retval = pam_start(SERVICE_NAME, user, &conv, &pamh);
	if(retval) {
		ERR("pam_start(%s) failed\n", user);
		return retval;
	}
	/* INFO("Setting PAM_RUSER='%s'\n", user); */
	retval = pam_set_item(pamh, PAM_RUSER, user);
	if(retval) {
		ERR("pam_set_item(PAM_RUSER): %s\n",
				pam_strerror(pamh, retval));
		goto out;
	}
	retval = pam_authenticate(pamh, 0);
	if(retval) {
		ERR("pam_authenticate: %s\n", pam_strerror(pamh, retval));
		goto out;
	}
	retval = pam_acct_mgmt(pamh, 0);
	if (retval) {
		ERR("pam_acct_mgmt: %s\n", pam_strerror(pamh, retval));
		goto out;
	}
	/* must be authenticated if this point was reached */
out:
	pam_end(pamh, PAM_SUCCESS);
	return retval;
}

#endif
