#include <assert.h>
#include <errno.h>
#include <glib.h>
#include <string.h>
#include <xtalk/debug.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_background.h>
#include <xpanel/xpw_hotkeys.h>
#include <xpanel/xpw_modal.h>
#include <xpanel/xpw_yesno.h>
#include <xpanel/xpw_modal.h>
#include <xpanel/xpw_ipaddr.h>
#include <xpanel/xpw_menu.h>
#include <xpanel/xpw_dialog.h>
#include <xpanel/xpw_editbox.h>
#include <xpanel/xpw_displaybox.h>
#include <xpanel/xpw_passwd.h>
#include <xpanel/kbd_glib.h>
#include <xpanel-auth.h>
#include <xpanel-ui.h>

#define	DBG_MASK	0x80

#define	MAX_INTERFACES	10
#define MAX_HARDWARE_PARAMS 2
#define XP_LCD_FLASHING_TIMEOUT 1200 //LCD flashing time on/off (msec)
#define XP_LCD_FAILURE_NOTIFICATION_DBOX_NAME "Failure"

#define	CALL_CHECK(ret, call) \
	if (ret >= 0) { \
		ret = (call); \
	}

#define	FAIL_CHECK(ret, target, call) \
	do { \
		ret = (call); \
		if (ret < 0) \
			goto target; \
	} while(0);

static struct xpanel_ui	*xpanel_ui = NULL;

/* forward declaractions */
static struct xpw_menu		top_menu;

static struct xpw_passwd	ui_passwd;

static struct xpw_menu_entry	me_status_menu;
static struct xpw_menu			status_menu;
static struct xpw_menu_entry		me_systemInfo;
static struct xpw_menu				systemInfo_menu;
static struct xpw_menu_entry			me_ip_status;
static struct xpw_dialog				ip_status_dialog;
static struct xpw_displaybox				ip_show[MAX_INTERFACES];

static struct xpw_menu_entry			me_hardware_status;
static struct xpw_dialog				hardware_status_dialog;
static struct xpw_displaybox				dbox_psu_sts;
static struct xpw_displaybox				dbox_raid_sts;

static struct xpw_menu_entry		me_twinstar_status;
static struct xpw_dialog			twinstar_status_dialog;
static struct xpw_displaybox				drbd_stat[2];
static struct xpw_menu_entry		me_rps_status;
static struct xpw_displaybox			rps_status_displaybox;
static struct xpw_menu_entry		me_pbx_menu;
static struct xpw_menu				pbx_menu;
static struct xpw_menu_entry			me_blink_port;
static struct xpw_menu					blink_port_menu;
static struct xpw_menu_entry				me_blink_extension;
static struct xpw_editbox					blink_extension_editbox;
static struct xpw_menu_entry				me_blink_span;
static struct xpw_editbox					blink_span_editbox;
static struct xpw_menu_entry				me_blink_channel;
static struct xpw_editbox					blink_channel_editbox;
static struct xpw_menu_entry				me_blink_xpd;
static struct xpw_editbox					blink_xpd_editbox;

static struct xpw_menu_entry			me_number_of_calls;
static struct xpw_displaybox				number_of_calls_displaybox;


static struct xpw_menu_entry	me_maintenance_menu;
static struct xpw_menu			maintenance_menu;

static struct xpw_menu_entry		me_asterisk_restart;
static struct xpw_menu					ast_restart_menu;
static struct xpw_menu_entry				me_ast_restart_yes;
static struct xpw_menu_notifier					ast_restart_yes_notifier;
static struct xpw_menu_entry				me_ast_restart_no;
static struct xpw_menu_notifier					ast_restart_no_notifier;

static struct xpw_menu_entry		me_dahdi_restart;
static struct xpw_menu					dahdi_restart_menu;
static struct xpw_menu_entry				me_dahdi_restart_yes;
static struct xpw_menu_notifier					dahdi_restart_yes_notifier;
static struct xpw_menu_entry				me_dahdi_restart_no;
static struct xpw_menu_notifier					dahdi_restart_no_notifier;

static struct xpw_menu_entry		me_reboot;
static struct xpw_menu					reboot_menu;
static struct xpw_menu_entry				me_reboot_yes;
static struct xpw_menu_notifier					reboot_yes_notifier;
static struct xpw_menu_entry				me_reboot_no;
static struct xpw_menu_notifier					reboot_no_notifier;

static struct xpw_menu_entry		me_poweroff;
static struct xpw_menu					poweroff_menu;
static struct xpw_menu_entry				me_poweroff_yes;
static struct xpw_menu_notifier					poweroff_yes_notifier;
static struct xpw_menu_entry				me_poweroff_no;
static struct xpw_menu_notifier					poweroff_no_notifier;


static struct xpw_menu_entry	me_settings_menu;
static struct xpw_menu			settings_menu;
static struct xpw_menu_entry		me_top_dhcp;
static struct xpw_menu				dhcp_menu;
static struct xpw_menu_entry				me_dhcp_interface[MAX_INTERFACES];
static struct xpw_menu_notifier				dhcp_interface[MAX_INTERFACES];
static struct xpw_menu_entry		me_top_manualnet;
static struct xpw_dialog			manualnet_dialog;
static struct xpw_ipaddr				manualnet_defaultgw;
static struct xpw_ipaddr				manualnet_dns1;
static struct xpw_ipaddr				manualnet_dns2;
static struct xpw_ipaddr				manualnet_ip[MAX_INTERFACES];
static struct xpw_ipaddr				manualnet_netmask[MAX_INTERFACES];
static struct xpw_yesno					manualnet_commit;
static struct xpw_modal					manualnet_modal;

static struct xpw_displaybox	failure_notification_displaybox;

void show_failure(gboolean on);

struct xpw *xpw_toplevel(void)
{
	return WIDGET(&top_menu);
}

static void got_yesno_answer(struct xpw *widget, enum keycodes key, void *data)
{
	struct xpw		*parent;
	struct xpw		*delegate;
	struct xpw_modal	*modal;
	struct xpw_yesno	*yesno;
	struct xpw_dialog	*dialog;
	int			answer;

	assert(widget->classid = XPW_MODAL);
	modal = (struct xpw_modal *)widget;
	parent = modal->parent;
	delegate = modal->delegate;
	assert(delegate->classid = XPW_YESNO);
	assert(parent->classid = XPW_DIALOG);
	dialog = (struct xpw_dialog *)parent;
	assert(dialog->myself.classid = XPW_DIALOG);
	yesno = (struct xpw_yesno *)delegate;
	answer = xpw_yesno_result(yesno);
	INFO("%s: %s\n", __func__, (answer) ? "YES" : "NO");
	if(answer) {
		DBG("COMMIT\n");
		modal_active(modal, 0);
		xpw_stack_pop();
		widget_notify(WIDGET(dialog), key);
	} /* else stay as we are */
}

static void do_asterisk_restart(struct xpw *widget, enum keycodes key, void *data)
{
	assert(xpanel_ui);
	INFO("%s %s\n", __func__, xpw_myname(widget));
	asterisk_restart(xpanel_ui->user_data);
}

static void do_dahdi_restart(struct xpw *widget, enum keycodes key, void *data)
{
	INFO("%s: %s\n", __func__, xpw_myname(widget));
	dahdi_restart(xpanel_ui->user_data);
}

static void do_reboot(struct xpw *widget, enum keycodes key, void *data)
{
	INFO("%s: %s\n", __func__, xpw_myname(widget));
	server_reboot(xpanel_ui->user_data);
}

static void do_poweroff (struct xpw *widget, enum keycodes key, void *data)
{
	INFO("%s: %s\n", __func__, xpw_myname(widget));
	server_poweroff(xpanel_ui->user_data);
}

static void blink_port(struct xpw *widget, enum keycodes key, void *data)
{
	// data - contains 'type'=ext, span, chan
	// The value (extension number, span number or channels number) must be 
	// taken from the edit box widget properties.
	struct xpw_editbox	*editbox = (struct xpw_editbox *)widget;
	INFO("%s: %s\n", __func__, xpw_myname(widget));
	blink_ports(xpanel_ui->user_data, (char *)data, (char *)editbox->contents);
}

gboolean get_fan_speed_ui(guchar *fan0, guchar *fan1, guchar *fan2) 
{
	return get_fan_speed( xpanel_ui->user_data, fan0, fan1, fan2);
}

gboolean lcd_flashing_cb(void *data)
{
	struct xpanel_ui	*ui = data;
	struct pcp_device	*pcpdev = ui->pcpdev;
	
	if(ui->lcd_flashing_sts) {
		// Continue flashing.
		if(pcp_getbacklightstatus(pcpdev))
			pcp_backlight(pcpdev, 0);
		else
			pcp_backlight(pcpdev, 1);
	} else {
		//Stop flashing
		pcp_backlight(pcpdev, 0);
	}
	return ui->lcd_flashing_sts;
}

gboolean psu_check_timeout_cb(void *data)
{
	struct xpanel_ui	*ui = data;
	struct pcp_device	*pcpdev = ui->pcpdev;
	int			psu_sts = PCP_PSU_STATUS_DISCONNECTED;
	guchar		action =0;
	int			rc = 0;

	rc = pcp_psu_read(pcpdev, &psu_sts);

	if(rc >= 0 && psu_sts != PCP_PSU_STATUS_DISCONNECTED && psu_sts != ui->psu_last_sts) {
		rc = psu_status_notification(ui->user_data, (guchar)psu_sts, &action);
		INFO("%s: PSU status=%d, prev. status=%d, action=%d\n", __FUNCTION__, psu_sts, ui->psu_last_sts, action);
		ui->psu_last_sts = psu_sts;

		if(action == 1) {
			ui->psu_failure_detected = TRUE;
			// Start buzzer
			pcp_buzzer(pcpdev, 1);
			
			//Show notification on the LCD. Here we put the text to xpanel_ui->failure_txt and 'push' the widget.
			//This text will be applied later in the failure_notification_populate() that will be called automatically.
			strncpy(ui->psu_failure_txt, "PSU", XPW_DISPLAYBOX_MAXWIDTH);
			show_failure(TRUE);

			// Start flashing
			if(!ui->lcd_flashing_sts) {
				ui->lcd_flashing_sts = TRUE;
				g_timeout_add(XP_LCD_FLASHING_TIMEOUT, lcd_flashing_cb, data);
			}
		} else if(action == 2) {
			ui->psu_failure_detected = FALSE;
			// Stop buzzer and flashing if possible.
			if(! ui->psu_failure_detected && ! ui->raid_failure_detected) {
				pcp_buzzer(pcpdev, 0);
				ui->lcd_flashing_sts = FALSE;
			}
			//Remove the notification message if necessary.
			ui->psu_failure_txt[0] = 0;
			show_failure(FALSE);
		}
	}

	return TRUE;
}

void raid_sts_cb(const char *txt)
{
	// This call-back function is called from xpanel-dbus.c after we've received a call to the 
	// xpanel_raid_info DBUS method.
	struct pcp_device	*pcpdev = xpanel_ui->pcpdev;
	char				tmp[XPW_DISPLAYBOX_MAXWIDTH+1];

	INFO("%s: text=%s\n", __func__, txt);
	
	if(txt && (strstr(txt, "OK") || strstr(txt, "N/A"))) {
		// RAID is ok
		xpanel_ui->raid_failure_detected = FALSE;
		// Stop buzzer and flashing if possible.
		if(! xpanel_ui->psu_failure_detected && ! xpanel_ui->raid_failure_detected) {
			pcp_buzzer(pcpdev, 0);
			xpanel_ui->lcd_flashing_sts = FALSE;
		}
		//Remove the notification message if necessary.
		xpanel_ui->raid_failure_txt[0] = 0;
		show_failure(FALSE);
	} else {
		xpanel_ui->raid_failure_detected = TRUE;
		// Start buzzer
		pcp_buzzer(pcpdev, 1);
		
		//Show notification on the LCD. Here we put the text to xpanel_ui->failure_txt and 'push' the widget.
		//This text will be applied later in the failure_notification_populate() that will be called automatically.
		strcpy(tmp, "HD ");
		strncat(tmp, txt, XPW_DISPLAYBOX_MAXWIDTH);
		tmp[XPW_DISPLAYBOX_MAXWIDTH-1] = 0;
		strncpy(xpanel_ui->raid_failure_txt, tmp, XPW_DISPLAYBOX_MAXWIDTH);
		show_failure(TRUE);

		// Start flashing
		if(!xpanel_ui->lcd_flashing_sts) {
			xpanel_ui->lcd_flashing_sts = TRUE;
			g_timeout_add(XP_LCD_FLASHING_TIMEOUT, lcd_flashing_cb, xpanel_ui);
		}
	}
}

static void dhcp_notify(struct xpw *widget, enum keycodes key, void *data)
{
	const char	*interface = data;
	dhcp_setter_t	dhcp_setter;

	assert(xpanel_ui);
	INFO("%s(%s): %s\n", __func__, interface, xpw_myname(widget));
	dhcp_setter = xpanel_ui->dhcp_setter;
	dhcp_setter(xpanel_ui->user_data, interface);
}

static void dhcp_cleanup(struct xpw *widget)
{
	int	i;
	GList	*curr;

	DBG("%s: %s\n", __func__, xpw_myname(widget));
	if(!xpanel_ui->netif_names)
		return;
	curr = g_list_first(xpanel_ui->netif_names);
	for(i = 0; curr && i < MAX_INTERFACES; i++, curr = g_list_next(curr)) {
		gpointer	name = curr->data;

		DBG("%s: %s -= %s\n", __func__, xpw_myname(widget), (char*)name);
		xpw_menu_entry_clear(&me_dhcp_interface[i]);
		xpw_set_callback(WIDGET(&dhcp_interface[i]), NULL, NULL);
	}
	xpw_menu_clear(&dhcp_menu);
	free_interface_names(xpanel_ui->user_data, xpanel_ui->netif_names);
	g_list_free(xpanel_ui->netif_names);
	xpanel_ui->netif_names = NULL;
}

static void dhcp_populate(struct xpw *widget)
{
	int		i;
	int		ret;
	struct xpw_menu	*dynamenu;
	GList		*curr;

	assert(widget->classid = XPW_MENU);
	assert(xpanel_ui);
	dynamenu = (struct xpw_menu *)widget;
	DBG("%s: %s\n", __func__, xpw_myname(widget));
	xpanel_ui->netif_names = get_interface_names(xpanel_ui->user_data);
	if(!xpanel_ui->netif_names) {
		ERR("%s: No network interfaces\n", __func__);
		return;
	}
	curr = g_list_first(xpanel_ui->netif_names);
	for(i = 0; curr && i < MAX_INTERFACES; i++, curr = g_list_next(curr)) {
		gpointer	name = curr->data;

		DBG("%s: %s += %s\n", __func__, xpw_myname(widget), (char*)name);
		FAIL_CHECK(ret, err, xpw_menu_notifier_construct(&dhcp_interface[i], "Sent", NULL));
		FAIL_CHECK(ret, err, xpw_menu_entry_construct(&me_dhcp_interface[i],
			WIDGET(&dhcp_interface[i]), name));
		xpw_set_callback(WIDGET(&dhcp_interface[i]), dhcp_notify, name);
		FAIL_CHECK(ret, err, xpw_menu_addentry(dynamenu, &me_dhcp_interface[i]));
	}
	return;
err:
	dhcp_cleanup(widget);
}

static int dhcp_settings_init(struct xpanel_ui *ui)
{
	int	ret = 0;

	FAIL_CHECK(ret, out, xpw_menu_construct(&dhcp_menu, "dhcp_menu"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_top_dhcp, WIDGET(&dhcp_menu), "DHCP"));
	FAIL_CHECK(ret, out, xpw_set_dynamic(WIDGET(&dhcp_menu), dhcp_populate, dhcp_cleanup));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&settings_menu, &me_top_dhcp));
	if (ret >= 0)
		ui->dhcp_setter = dhcp_setter;
out:
	return ret;
}

static void ip_status_cleanup(struct xpw *widget)
{
	struct xpw_dialog	*dialog;
	int			i;
	GList		*curr;

	INFO("%s\n", __func__);

	dialog = (struct xpw_dialog *) widget;

	if(xpanel_ui->netif_names) {
		curr = g_list_first(xpanel_ui->netif_names);
		for(i = 0; curr && i < MAX_INTERFACES; i++, curr = g_list_next(curr)) {
			gpointer	name = curr->data;

			DBG("%s: %s -= %s\n", __func__, xpw_myname(widget), (char*)name);
			xpw_dialog_delwidget(dialog, WIDGET(&ip_show[i]));
			g_free(name);
		}
		free_interface_names(xpanel_ui->user_data, xpanel_ui->netif_names);
		xpanel_ui->netif_names = NULL;
	}
}

static void ip_status_populate(struct xpw *widget)
{
	int			ret = 0;
	struct xpw_dialog	*dialog;
	int			i;
	GList		*curr;

	INFO("%s\n", __func__);

	dialog = (struct xpw_dialog *) widget;

	xpanel_ui->netif_names = get_interface_names(xpanel_ui->user_data);
	if(!xpanel_ui->netif_names) {
		ERR("%s: No network interfaces\n", __func__);
		goto err;
	}

	curr = g_list_first(xpanel_ui->netif_names);
	for(i = 0; curr && i < MAX_INTERFACES; i++, curr = g_list_next(curr)) {
		gpointer	name = curr->data;
		char		tmpbuf[BUFSIZ];
		char		*ipaddr = NULL;
		char		*netmask = NULL;

		if(!get_interface_settings(xpanel_ui->user_data, name, &ipaddr, &netmask)) {
			ERR("Failed getting IP+Netmask of %s. Setting defaults\n", (char*)name);
			ipaddr = strdup("0.0.0.0");
		}
		DBG("%s: %s += %s\n", __func__, xpw_myname(widget), (char*)name);
		snprintf(tmpbuf, sizeof(tmpbuf), "%s IP", (char*)name);
		CALL_CHECK(ret, xpw_displaybox_construct(&ip_show[i], tmpbuf));
		CALL_CHECK(ret, xpw_displaybox_set( &ip_show[i], ipaddr));
		CALL_CHECK(ret, xpw_dialog_addwidget(dialog, WIDGET(&ip_show[i])));
		g_free(ipaddr);
		if(ret < 0) 
			goto err;
	}
	return;
err:
	ERR("Failed populating ip_status\n");
	ip_status_cleanup(widget);
}

static void hardware_status_cleanup(struct xpw *widget)
{
	struct xpw_dialog	*dialog;

	INFO("%s\n", __func__);

	dialog = (struct xpw_dialog *) widget;

	xpw_dialog_delwidget(dialog, WIDGET(&dbox_psu_sts));
	xpw_dialog_delwidget(dialog, WIDGET(&dbox_raid_sts));
}

static void hardware_status_populate(struct xpw *widget)
{
	struct xpw_dialog	*dialog;
	int			ret = 0,
				rc =0;
	int			psu_sts = PCP_PSU_STATUS_DISCONNECTED;
	
	char		*psu_sts_txt = NULL;
	char		*raid_sts_txt = NULL;
	char		tmp[XPW_DISPLAYBOX_MAXWIDTH];

	dialog = (struct xpw_dialog *)widget;

	INFO("%s\n", __func__);

	// Get the PSU status
	rc = pcp_psu_read(xpanel_ui->pcpdev, &psu_sts);

	if(rc >= 0) {
		switch (psu_sts) {
			case PCP_PSU_STATUS_OFF:
				psu_sts_txt = strdup("  Failure!");
				break;
			case PCP_PSU_STATUS_ON:
				psu_sts_txt = strdup("  OK");
				break;
			default:
				psu_sts_txt = strdup("  N/A");
		}
	} else 
		psu_sts_txt = strdup("  N/A");
	
	// Get the RAID status via DBUS and build the message.
	tmp[0] = ' ';
	tmp[1] = 0;
	if(!get_raid_status(xpanel_ui->user_data, &raid_sts_txt)) {
		ERR("Failed to get the RAID status.\n");
		strcat(tmp, "N/A");
	} else {
		if( strstr(raid_sts_txt, "OK") == NULL && strstr(raid_sts_txt, "N/A") == NULL) {
			strcat(tmp, "HD ");
			strncat(tmp, raid_sts_txt, XPW_DISPLAYBOX_MAXWIDTH - 1);
			strcat(tmp, " faulty");
		} else {
			strncat(tmp, raid_sts_txt, XPW_DISPLAYBOX_MAXWIDTH - 1);
		}
		
		tmp[XPW_DISPLAYBOX_MAXWIDTH-1] = 0;
	}

	FAIL_CHECK(ret, out, xpw_displaybox_construct(&dbox_psu_sts, "PSU status"));
	FAIL_CHECK(ret, out, xpw_displaybox_set( &dbox_psu_sts, psu_sts_txt));
	FAIL_CHECK(ret, out, xpw_dialog_addwidget(dialog, WIDGET(&dbox_psu_sts)));
	
	FAIL_CHECK(ret, out, xpw_displaybox_construct(&dbox_raid_sts, "RAID status"));
	FAIL_CHECK(ret, out, xpw_displaybox_set( &dbox_raid_sts, tmp));
	FAIL_CHECK(ret, out, xpw_dialog_addwidget(dialog, WIDGET(&dbox_raid_sts)));
out:
	if (psu_sts_txt)
		g_free(psu_sts_txt);
	if (raid_sts_txt)
		g_free(raid_sts_txt);
	if(ret < 0) {
		ERR("Failed populating hardware_status_dialog\n");
		hardware_status_cleanup(widget);
	} 
	return;
}

static void failure_notification_populate(struct xpw *widget) {
	struct xpw_displaybox *dbox = (struct xpw_displaybox *) widget;
	int 	eol = XPW_DISPLAYBOX_MAXWIDTH;
	
	INFO("%s\n", __func__);
	char	notification_text[XPW_DISPLAYBOX_MAXWIDTH*2+5];
	
	notification_text[0] = ' ';
	notification_text[1] = 0;
	
	strncat(notification_text, xpanel_ui->psu_failure_txt, XPW_DISPLAYBOX_MAXWIDTH);
	if(strlen(notification_text) > 1 && xpanel_ui->raid_failure_txt[0] != 0) 
		strcat(notification_text, ", ");
	strncat(notification_text, xpanel_ui->raid_failure_txt, XPW_DISPLAYBOX_MAXWIDTH);
	if(strlen(notification_text) < XPW_DISPLAYBOX_MAXWIDTH) 
		eol = strlen(notification_text);
	notification_text[eol] = 0; 

	xpw_displaybox_set(dbox, notification_text);
}

static void failure_notification_cleanup(struct xpw *widget) {
	struct xpw_displaybox *dbox = (struct xpw_displaybox *) widget;

	INFO("%s\n", __func__);
	
	xpanel_ui->psu_failure_txt[0] = 0;
	xpanel_ui->raid_failure_txt[0] = 0;
	xpw_displaybox_set(dbox, "");
}

void show_failure(gboolean on)
{
	struct xpw		*widget;
	gboolean		displayed = FALSE;

	INFO("%s show: %d\n", __func__, on);
	
	widget = xpw_current();

	if (strcmp(xpw_myname(widget), XP_LCD_FAILURE_NOTIFICATION_DBOX_NAME) == 0)
		displayed = TRUE;
	
	if (displayed && on) {
		// Redraw text
		failure_notification_populate(widget);
		widget_redraw(xpanel_ui->screen, WIDGET(widget));
	} else 	if(displayed && !on) {
		// Remove the notification if it is possible or just update the text.
		if( xpanel_ui->psu_failure_txt[0] == 0 && xpanel_ui->raid_failure_txt[0] == 0) {
			xpw_stack_pop();
			widget = xpw_current();
			widget_redraw(xpanel_ui->screen, WIDGET(widget));
		} else {
			// Update the notification text only.
			failure_notification_populate(widget);
			widget_redraw(xpanel_ui->screen, WIDGET(widget));
		}
	} else if (!displayed && on){
		// Create notification message. Note, that failure_notification_populate()
		// will be called automatically.
		xpw_stack_push(WIDGET(&failure_notification_displaybox));
		widget_redraw(xpanel_ui->screen, WIDGET(&failure_notification_displaybox));
	}
}

static void twinstar_status_cleanup(struct xpw *widget)
{
	struct xpw_dialog	*dialog;

	INFO("%s\n", __func__);

	dialog = (struct xpw_dialog *) widget;

	xpw_dialog_delwidget(dialog, WIDGET(&drbd_stat[0]));
	xpw_dialog_delwidget(dialog, WIDGET(&drbd_stat[1]));
}

static void twinstar_status_populate(struct xpw *widget)
{
	struct xpw_dialog	*dialog;
	int			ret = 0;
	char			*drbd_role = NULL;
	char			*drbd_cs = NULL;

	dialog = (struct xpw_dialog *)widget;

	// Get the DRBD statuses via DBUS
	if(!get_twinstar_status(xpanel_ui->user_data, &drbd_role, &drbd_cs)) {
		ERR("Failed to get the DRBD status.\n");
		drbd_role = strdup("N/A");
		drbd_cs = strdup("N/A");
	}

	FAIL_CHECK(ret, out, xpw_displaybox_construct(&drbd_stat[0], "DRBD role"));
	FAIL_CHECK(ret, out, xpw_displaybox_set( &drbd_stat[0], drbd_role));
	FAIL_CHECK(ret, out, xpw_dialog_addwidget(dialog, WIDGET(&drbd_stat[0])));
	FAIL_CHECK(ret, out, xpw_displaybox_construct(&drbd_stat[1], "DRBD connect"));
	FAIL_CHECK(ret, out, xpw_displaybox_set( &drbd_stat[1], drbd_cs));
	FAIL_CHECK(ret, out, xpw_dialog_addwidget(dialog, WIDGET(&drbd_stat[1])));
out:
	if (drbd_role)
		g_free(drbd_role);
	if (drbd_cs)
		g_free(drbd_cs);
	if(ret < 0) {
		ERR("Failed populating twinstar_status_dialog\n");
		twinstar_status_cleanup(widget);
	} 
	return;
}

static void number_of_calls_cleanup(struct xpw *widget)
{
	struct xpw_displaybox	*num_calls;
	num_calls = (struct xpw_displaybox *)widget;

	memset( num_calls->contents, 0, XPW_DISPLAYBOX_MAXWIDTH);
}

static void number_of_calls_populate(struct xpw *widget)
{
	char  		*num_calls = NULL;

	struct xpw_displaybox	*disp_box;
	disp_box = (struct xpw_displaybox *)widget;

	if( get_num_calls(xpanel_ui->user_data, &num_calls))
		strncpy( disp_box->contents, num_calls, XPW_DISPLAYBOX_MAXWIDTH);
	else
		strncpy( disp_box->contents, "N/A", XPW_DISPLAYBOX_MAXWIDTH);

	g_free(num_calls);
}


static void rps_status_cleanup(struct xpw *widget)
{
	struct xpw_displaybox	*rbs_sts;
	rbs_sts = (struct xpw_displaybox *)widget;

	memset( rbs_sts->contents, 0, XPW_DISPLAYBOX_MAXWIDTH);
}

static void rps_status_populate(struct xpw *widget)
{
	GList		*names_lst = NULL;
	GList		*device = NULL;
	gint		mask = 0;
	int  		ps_voltage[4];
	int		len = 0;
	int		i;
	struct xpw_displaybox	*rbs_sts;

#define	FILL_RBS_STS(rbs_sts, offset, txt) \
		(offset) += \
			snprintf((rbs_sts)->contents + (offset), \
				XPW_DISPLAYBOX_MAXWIDTH - (offset), \
				txt)

	rbs_sts = (struct xpw_displaybox *)widget;

	names_lst = get_rps_devices(xpanel_ui->user_data);
	if(!names_lst) {
		FILL_RBS_STS(rbs_sts, len, "Not connected");
		goto out;
	}
	//[TBD] Add support for several RPS connected.
	// Meantime show status of the first RPS in the list only.
	device = g_list_first(names_lst);
	if (! get_rps_mask(xpanel_ui->user_data, (char *)device->data, &mask)) {
		FILL_RBS_STS(rbs_sts, len, "ERR: Mask");
		goto out;
	}
	if (! get_rps_voltage(xpanel_ui->user_data, (char *)device->data, ps_voltage)) {
		FILL_RBS_STS(rbs_sts, len, "ERR: Voltage");
		goto out;
	}
	for(i = 0; i < 4; i++) {
		if( mask & (0x1<< i) )
			if(ps_voltage[i] < 110)
				FILL_RBS_STS(rbs_sts, len, "OFF ");
			else
				FILL_RBS_STS(rbs_sts, len, "ON  ");
		else
			FILL_RBS_STS(rbs_sts, len, "N/A ");
	}
out:
	g_list_free(names_lst);
#undef	FILL_RBS_STS
}


static int build_systemInfo_menu()
{
	int	ret = 0;

	FAIL_CHECK(ret, out, xpw_dialog_construct(&ip_status_dialog, "IP address:"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_ip_status, WIDGET(&ip_status_dialog), "IP address"));
	FAIL_CHECK(ret, out, xpw_set_dynamic(WIDGET(&ip_status_dialog), ip_status_populate, ip_status_cleanup));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&systemInfo_menu, &me_ip_status));

	FAIL_CHECK(ret, out, xpw_dialog_construct(&hardware_status_dialog, "HW status:"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_hardware_status, WIDGET(&hardware_status_dialog), "HW status"));
	FAIL_CHECK(ret, out, xpw_set_dynamic(WIDGET(&hardware_status_dialog), hardware_status_populate, hardware_status_cleanup));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&systemInfo_menu, &me_hardware_status));
out:
	return ret;
}

static int build_pbx_menu()
{
	int	ret = 0;

	// Add entries to the pbx_menu
	FAIL_CHECK(ret, out, xpw_menu_construct(&blink_port_menu, "blnk_port_m"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_blink_port, WIDGET(&blink_port_menu), "Blink Port"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&pbx_menu, &me_blink_port));

	FAIL_CHECK(ret, out, xpw_displaybox_construct(&number_of_calls_displaybox, "ActiveCalls"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_number_of_calls, WIDGET(&number_of_calls_displaybox), "Active Calls"));
	FAIL_CHECK(ret, out, xpw_set_dynamic(WIDGET(&number_of_calls_displaybox), number_of_calls_populate, number_of_calls_cleanup));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&pbx_menu, &me_number_of_calls));

	// Build the blink_port_menu
	FAIL_CHECK(ret, out, xpw_editbox_construct(&blink_extension_editbox,"Extension"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_blink_extension, WIDGET(&blink_extension_editbox), "Extension"));
	FAIL_CHECK(ret, out, xpw_set_callback(WIDGET(&blink_extension_editbox), blink_port, "ext"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&blink_port_menu, &me_blink_extension));

	FAIL_CHECK(ret, out, xpw_editbox_construct(&blink_channel_editbox,"Channel")); 
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_blink_channel, WIDGET(&blink_channel_editbox), "Channel"));
	FAIL_CHECK(ret, out, xpw_set_callback(WIDGET(&blink_channel_editbox), blink_port, "chan"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&blink_port_menu, &me_blink_channel));

	FAIL_CHECK(ret, out, xpw_editbox_construct(&blink_xpd_editbox,"XPD")); 
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_blink_xpd, WIDGET(&blink_xpd_editbox), "XPD"));
	FAIL_CHECK(ret, out, xpw_set_callback(WIDGET(&blink_xpd_editbox), blink_port, "xpd"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&blink_port_menu, &me_blink_xpd));

	FAIL_CHECK(ret, out, xpw_editbox_construct(&blink_span_editbox,"SPAN"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_blink_span, WIDGET(&blink_span_editbox), "SPAN"));
	FAIL_CHECK(ret, out, xpw_set_callback(WIDGET(&blink_span_editbox), blink_port, "span"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&blink_port_menu, &me_blink_span));
out:
	return ret;
}

static void manualnet_update(struct xpw *widget, enum keycodes key, void *data)
{
// *data - pointer to XPanelDaemon *daemon. XPanelDaemon contains pointer 'XPanelDaemonPrivate *priv'
// 		   that in turn contains pointers to d-bus related objects.
// 
	struct xpw_dialog	*xpw_dialog;
	struct xpw		*entry;
	struct xpw_ipaddr	*xpw_ipaddr;
#define	IP_ITEMS				10
#define MAX_INTREFACES			5
#define MAX_INTEFACE_DESC_LEN	50+1 
	char			str_ipdata[IP_ITEMS][MAX_INTEFACE_DESC_LEN];
	char			**ip_interfaces;
	char 			delimiters[] = " :";
	gchar			* tmp_str,
				*if_name,
				*if_par;
	int			i,
				curr_interface =0;

	assert(widget->classid == XPW_DIALOG);
	xpw_dialog = (struct xpw_dialog *)widget;

	ip_interfaces = g_new( char *, MAX_INTREFACES);
	for(i=0; i < MAX_INTREFACES; i++)
	  ip_interfaces[i] = NULL;

	for(i = 0; i < IP_ITEMS; i++) {
		entry = xpw_dialog->entries[i];
		if(!entry) {
			str_ipdata[i][0] = '\0';
			continue;
		}
		assert(entry->classid == XPW_IPADDR);
		xpw_ipaddr = (struct xpw_ipaddr *)entry;
		xpw_ipaddr_result(xpw_ipaddr, str_ipdata[i], sizeof(str_ipdata[i]));

		if(i > 2) {
			INFO("%s -> '%s'\n", xpw_ipaddr->prompt, str_ipdata[i]);
			//take first word from xpw_ipaddr->prompt  eth0 IP:  eth0 Netmask:
			tmp_str = g_strdup( xpw_ipaddr->prompt );
			if_name = strtok( tmp_str, delimiters );
			if_par = strtok( NULL, delimiters );
			if ( ! strcmp( if_par, "IP")) {
				if ( !ip_interfaces[ curr_interface] )
					ip_interfaces[ curr_interface] = g_strnfill( MAX_INTEFACE_DESC_LEN, ' ' );
				g_strlcpy( ip_interfaces[ curr_interface],if_name, MAX_INTEFACE_DESC_LEN);
				g_strlcat( ip_interfaces[ curr_interface], "-", MAX_INTEFACE_DESC_LEN);
				g_strlcat( ip_interfaces[ curr_interface], str_ipdata[i], MAX_INTEFACE_DESC_LEN);
				g_strlcat( ip_interfaces[ curr_interface], "-", MAX_INTEFACE_DESC_LEN);
			}
			else {
				g_strlcat( ip_interfaces[ curr_interface], str_ipdata[i], MAX_INTEFACE_DESC_LEN);
				curr_interface++;			
			}
			g_free(tmp_str);
		}
	}
	INFO("FINALLY: GW=\"%s\" DNS1=\"%s\" DNS2=\"%s\"\n",
		str_ipdata[0], str_ipdata[1], str_ipdata[2]);

	apply_manualnet_update(data, str_ipdata[0], str_ipdata[1], str_ipdata[2], (const char **)ip_interfaces);

	g_strfreev( ip_interfaces);
}


static void manualnet_cleanup(struct xpw *widget)
{
	struct xpw_dialog	*dialog;
	GList			*curr;
	int			i;

	assert(widget->classid = XPW_DIALOG);
	dialog = (struct xpw_dialog *)widget;

	free_network_settings(xpanel_ui->user_data, &xpanel_ui->network_settings);
	xpw_dialog_delwidget(dialog, WIDGET(&manualnet_defaultgw));
	xpw_dialog_delwidget(dialog, WIDGET(&manualnet_dns1));
	xpw_dialog_delwidget(dialog, WIDGET(&manualnet_dns2));

	if(xpanel_ui->netif_names) {
		curr = g_list_first(xpanel_ui->netif_names);
		for(i = 0; curr && i < MAX_INTERFACES; i++, curr = g_list_next(curr)) {
			gpointer	name = curr->data;

			DBG("%s: %s -= %s\n", __func__, xpw_myname(widget), (char*)name);
			xpw_dialog_delwidget(dialog, WIDGET(&manualnet_ip[i]));
			xpw_set_callback(WIDGET(&manualnet_ip[i]), NULL, NULL);
			xpw_dialog_delwidget(dialog, WIDGET(&manualnet_netmask[i]));
			xpw_set_callback(WIDGET(&manualnet_netmask[i]), NULL, NULL);
			g_free(name);
		}
		free_interface_names(xpanel_ui->user_data, xpanel_ui->netif_names);
		xpanel_ui->netif_names = NULL;
	}
}

static void manualnet_populate(struct xpw *widget)
{
	struct network_settings	*ns;
	struct xpw_dialog	*dialog;
	int					ret;
	struct in_addr		addr;
	int					i;
	GList				*curr;

	INFO("%s\n", __func__);

	assert(widget->classid = XPW_DIALOG);
	dialog = (struct xpw_dialog *)widget;
	ns = &xpanel_ui->network_settings;
	xpanel_ui->netif_names = get_interface_names(xpanel_ui->user_data);
	if(!xpanel_ui->netif_names) {
		ERR("%s: No network interfaces\n", __func__);
		goto err;
	}

	get_network_settings(xpanel_ui->user_data, ns);
	INFO("Got: DEFAULTGW=%s DNS1=%s DNS2=%s\n",
		ns->defaultgw,
		ns->dns1,
		ns->dns2);
	if(!ns->defaultgw || !inet_pton(AF_INET, ns->defaultgw, &addr))
		addr.s_addr = htonl(0L);
	FAIL_CHECK(ret, err, xpw_ipaddr_construct(&manualnet_defaultgw, "Enter Gateway", addr.s_addr));
	FAIL_CHECK(ret, err, xpw_dialog_addwidget(dialog, WIDGET(&manualnet_defaultgw)));
	if(!ns->dns1 || !inet_pton(AF_INET, ns->dns1, &addr))
		addr.s_addr = htonl(0L);
	FAIL_CHECK(ret, err, xpw_ipaddr_construct(&manualnet_dns1, "Enter DNS1", addr.s_addr));
	FAIL_CHECK(ret, err, xpw_dialog_addwidget(dialog, WIDGET(&manualnet_dns1)));
	if(!ns->dns2 || !inet_pton(AF_INET, ns->dns2, &addr))
		addr.s_addr = htonl(0L);
	FAIL_CHECK(ret, err, xpw_ipaddr_construct(&manualnet_dns2, "Enter DNS2", addr.s_addr));
	FAIL_CHECK(ret, err, xpw_dialog_addwidget(dialog, WIDGET(&manualnet_dns2)));

	curr = g_list_first(xpanel_ui->netif_names);
	for(i = 0; curr && i < MAX_INTERFACES; i++, curr = g_list_next(curr)) {
		gpointer	name = curr->data;
		char		tmpbuf[BUFSIZ];
		char		*ipaddr = NULL;
		char		*netmask = NULL;
		gboolean	bErr = TRUE;

		if(!get_interface_settings(xpanel_ui->user_data, name, &ipaddr, &netmask)) {
			ERR("Failed getting IP+Netmask of %s. Setting defaults\n", (char*)name);
			ipaddr = strdup("0.0.0.0");
			netmask = strdup("0.0.0.0");
		}
		DBG("%s: %s += %s\n", __func__, xpw_myname(widget), (char*)name);
		if(!inet_pton(AF_INET, ipaddr, &addr)) {
			ERR("Bad IP address '%s'\n", ipaddr);
			bErr = FALSE;
		}

		if(bErr) {
			snprintf(tmpbuf, sizeof(tmpbuf), "%s IP", (char*)name);
			ret = xpw_ipaddr_construct(&manualnet_ip[i], tmpbuf, addr.s_addr);
			if(ret < 0)
				bErr = FALSE;
			else {
				ret = xpw_dialog_addwidget(dialog, WIDGET(&manualnet_ip[i]));
				if(ret < 0)
					bErr = FALSE;
			}
		}

		if(bErr) {
			if(!inet_pton(AF_INET, netmask, &addr)) {
				ERR("Bad IP address '%s'\n", netmask);
				bErr = FALSE;
			} else {
			  snprintf(tmpbuf, sizeof(tmpbuf), "%s Netmask", (char*)name);
			  ret = xpw_ipaddr_construct(&manualnet_netmask[i], tmpbuf, addr.s_addr);
			  if(ret < 0)
				  bErr = FALSE;
			}
		}
		if(bErr) {
			ret = xpw_dialog_addwidget(dialog, WIDGET(&manualnet_netmask[i]));
			if(ret < 0)
				bErr = FALSE;
		}
		g_free(ipaddr);
		g_free(netmask);
		if(!bErr)
			goto err;
	}
	/* Add confirmation */
	FAIL_CHECK(ret, err, xpw_yesno_construct(&manualnet_commit, "Save Network", "Yes", "No", 1));
	FAIL_CHECK(ret, err, xpw_set_modal(WIDGET(dialog), &manualnet_modal, WIDGET(&manualnet_commit)));
	xpw_set_callback(WIDGET(&manualnet_modal), got_yesno_answer, NULL);
	xpw_set_callback(WIDGET(&manualnet_dialog), manualnet_update, xpanel_ui->user_data);
	return;
err:
	ERR("Failed populating manualnet\n");
	manualnet_cleanup(widget);
}

static int manualnet_settings_init(void)
{
	int	ret = 0;

	FAIL_CHECK(ret, out, xpw_dialog_construct(&manualnet_dialog, "Manual Net"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_top_manualnet, WIDGET(&manualnet_dialog), "Manual Net"));
	FAIL_CHECK(ret, out, xpw_set_dynamic(WIDGET(&manualnet_dialog), manualnet_populate, manualnet_cleanup));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&settings_menu, &me_top_manualnet));
out:
	return ret;
}

static int build_top_level_menu()
{
	int	ret = 0;

	FAIL_CHECK(ret, out, xpw_menu_construct(&top_menu, "top_menu"));
	FAIL_CHECK(ret, out, xpw_menu_construct(&status_menu, "status_menu"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_status_menu, WIDGET(&status_menu), "Status"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&top_menu, &me_status_menu));

	FAIL_CHECK(ret, out, xpw_menu_construct(&maintenance_menu, "maintenance_menu"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_maintenance_menu, WIDGET(&maintenance_menu), "Maintenance"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&top_menu, &me_maintenance_menu));

	FAIL_CHECK(ret, out, xpw_menu_construct(&settings_menu, "settings_menu"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_settings_menu, WIDGET(&settings_menu), "Settings"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&top_menu, &me_settings_menu));
out:
	return ret;
}

static int build_status_menu()
{
	int	ret = 0;

	FAIL_CHECK(ret, out, xpw_menu_construct(&systemInfo_menu, "systemInfo_menu"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_systemInfo, WIDGET(&systemInfo_menu), "SystemInfo"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&status_menu, &me_systemInfo));

	FAIL_CHECK(ret, out, build_systemInfo_menu());

	FAIL_CHECK(ret, out, xpw_menu_construct(&pbx_menu, "pbx_menu"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_pbx_menu, WIDGET(&pbx_menu), "PBX Info"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&status_menu, &me_pbx_menu));

	FAIL_CHECK(ret, out, build_pbx_menu());

	FAIL_CHECK(ret, out, xpw_dialog_construct(&twinstar_status_dialog, "Server State:"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_twinstar_status, WIDGET(&twinstar_status_dialog), "TwinStarStat"));
	FAIL_CHECK(ret, out, xpw_set_dynamic(WIDGET(&twinstar_status_dialog), twinstar_status_populate, twinstar_status_cleanup));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&status_menu, &me_twinstar_status));

	FAIL_CHECK(ret, out, xpw_displaybox_construct(&rps_status_displaybox, "RPS Status"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_rps_status, WIDGET(&rps_status_displaybox), "RPS Status"));
	FAIL_CHECK(ret, out, xpw_set_dynamic(WIDGET(&rps_status_displaybox), rps_status_populate, rps_status_cleanup));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&status_menu, &me_rps_status));
out:
	return ret;
}


static int build_maintenance_menu()
{
	int	ret = 0;

	// Asterisk restart
	FAIL_CHECK(ret, out, xpw_menu_construct(&ast_restart_menu, "ast_rest_m"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_asterisk_restart, WIDGET(&ast_restart_menu), "AstrskRstart"));

	FAIL_CHECK(ret, out, xpw_menu_notifier_construct(&ast_restart_yes_notifier, "Done", NULL));
	FAIL_CHECK(ret, out, xpw_set_callback(WIDGET(&ast_restart_yes_notifier), do_asterisk_restart, "AstrskRstart"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_ast_restart_yes, WIDGET(&ast_restart_yes_notifier), "Restart"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&ast_restart_menu, &me_ast_restart_yes));

	FAIL_CHECK(ret, out, xpw_menu_notifier_construct(&ast_restart_no_notifier, "Canceled", NULL));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_ast_restart_no, WIDGET(&ast_restart_no_notifier), "Cancel"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&ast_restart_menu, &me_ast_restart_no));

	FAIL_CHECK(ret, out, xpw_menu_addentry(&maintenance_menu, &me_asterisk_restart));

	// DAHDI restart
	FAIL_CHECK(ret, out, xpw_menu_construct(&dahdi_restart_menu, "dahdi_rest_m"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_dahdi_restart, WIDGET(&dahdi_restart_menu), "DahdiRstart"));

	FAIL_CHECK(ret, out, xpw_menu_notifier_construct(&dahdi_restart_yes_notifier, "Done", NULL));
	FAIL_CHECK(ret, out, xpw_set_callback(WIDGET(&dahdi_restart_yes_notifier), do_dahdi_restart, "DahdiRstart"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_dahdi_restart_yes, WIDGET(&dahdi_restart_yes_notifier), "Confirm"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&dahdi_restart_menu, &me_dahdi_restart_yes));

	FAIL_CHECK(ret, out, xpw_menu_notifier_construct(&dahdi_restart_no_notifier, "Canceled", NULL));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_dahdi_restart_no, WIDGET(&dahdi_restart_no_notifier), "Cancel"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&dahdi_restart_menu, &me_dahdi_restart_no));

	FAIL_CHECK(ret, out, xpw_menu_addentry(&maintenance_menu, &me_dahdi_restart));

	// Reboot
	FAIL_CHECK(ret, out, xpw_menu_construct(&reboot_menu, "reboot_m"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_reboot, WIDGET(&reboot_menu), "Reboot"));

	FAIL_CHECK(ret, out, xpw_menu_notifier_construct(&reboot_yes_notifier, "Done", NULL));
	FAIL_CHECK(ret, out, xpw_set_callback(WIDGET(&reboot_yes_notifier), do_reboot, "Reboot"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_reboot_yes, WIDGET(&reboot_yes_notifier), "Confirm"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&reboot_menu, &me_reboot_yes));

	FAIL_CHECK(ret, out, xpw_menu_notifier_construct(&reboot_no_notifier, "Canceled", NULL));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_reboot_no, WIDGET(&reboot_no_notifier), "Cancel"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&reboot_menu, &me_reboot_no));

	FAIL_CHECK(ret, out, xpw_menu_addentry(&maintenance_menu, &me_reboot));

	// Poweroff
	FAIL_CHECK(ret, out, xpw_menu_construct(&poweroff_menu, "poweroff_m"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_poweroff, WIDGET(&poweroff_menu), "Poweroff"));

	FAIL_CHECK(ret, out, xpw_menu_notifier_construct(&poweroff_yes_notifier, "Done", NULL));
	FAIL_CHECK(ret, out, xpw_set_callback(WIDGET(&poweroff_yes_notifier), do_poweroff, "Poweroff"));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_poweroff_yes, WIDGET(&poweroff_yes_notifier), "Confirm"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&poweroff_menu, &me_poweroff_yes));

	FAIL_CHECK(ret, out, xpw_menu_notifier_construct(&poweroff_no_notifier, "Canceled", NULL));
	FAIL_CHECK(ret, out, xpw_menu_entry_construct(&me_poweroff_no, WIDGET(&poweroff_no_notifier), "Cancel"));
	FAIL_CHECK(ret, out, xpw_menu_addentry(&poweroff_menu, &me_poweroff_no));

	FAIL_CHECK(ret, out, xpw_menu_addentry(&maintenance_menu, &me_poweroff));
out:
	return ret;
}

static int build_settings_menu(struct xpanel_ui *ui)
{
	int	ret = 0;

	FAIL_CHECK(ret, out, dhcp_settings_init(ui));
	FAIL_CHECK(ret, out, manualnet_settings_init());
out:
	return ret;
}

void stop_flashing()
{
	xpanel_ui->lcd_flashing_sts = FALSE;
}

struct xpanel_ui *xpanel_ui_new(struct pcp_device *pcpdev, struct xpw_screen *screen,
								 int lock_time, int psu_time, 
								 const char *login_name, const char *logo)
{
	struct xpw_background	*background = NULL;
	struct xpanel_ui	*ui;
	int			ret;

	if((ui = malloc(sizeof(struct xpanel_ui))) == NULL)
		return NULL;

	memset(ui, 0, sizeof(*ui));
	if (!login_name || login_name[0] == '\0') {
		INFO("No login screen\n");
		ui->login_name = NULL;
	} else {
		INFO("Login as '%s'\n", login_name);
		ui->login_name = strdup(login_name);
	}
	ui->screen_lock_facade = XPW_PASSWD_NO_FACADE;
	ui->pcpdev = pcpdev;
	ui->screen = screen;
	ui->screen_lock_time = lock_time;
	ui->psu_check_time = psu_time;
	ui->psu_last_sts = PCP_PSU_STATUS_DISCONNECTED;
	ui->psu_failure_detected = FALSE;
	ui->raid_failure_detected = FALSE;
	ui->psu_failure_txt[0] = 0;
	ui->raid_failure_txt[0] = 0;
	ui->toplevel = WIDGET(&top_menu);
	ret = xpw_init(ui->toplevel);
	FAIL_CHECK(ret, err, xpw_background_startup(ui->toplevel, logo));
	background = xpw_background_get();
	if (!background) {
		ret = -EINVAL;
		goto err;
	}

	// Build the top level menu.
	FAIL_CHECK(ret, err, build_top_level_menu());

	// Build the "Status" menu
	FAIL_CHECK(ret, err, build_status_menu(ui));

	// Build the "Maintenance" menu
	FAIL_CHECK(ret, err, build_maintenance_menu());

	// Build the "Settings" menu
	FAIL_CHECK(ret, err, build_settings_menu(ui));

	// Create a standalone display box for failure notifications.
	FAIL_CHECK(ret, err, xpw_displaybox_construct(&failure_notification_displaybox, XP_LCD_FAILURE_NOTIFICATION_DBOX_NAME));
	FAIL_CHECK(ret, err, xpw_set_dynamic(WIDGET(&failure_notification_displaybox), failure_notification_populate,
												failure_notification_cleanup));
	
	// Init hot keys
	FAIL_CHECK(ret, err, xpw_hotkey_set(KEYCODE_IPADDR, WIDGET(&ip_status_dialog), NULL));
	FAIL_CHECK(ret, err, xpw_hotkey_set(KEYCODE_DOWN, WIDGET(&poweroff_menu), NULL));

	xpw_stack_push(WIDGET(background));
	widget_redraw(screen, WIDGET(background));
	xpanel_ui = ui;
	return ui;
err:
	if(ui)
		free(ui);
	return NULL;
}

void xpanel_ui_bind(struct xpanel_ui *ui, void *data)
{
	assert(ui);
	ui->user_data = data;
}

void xpanel_ui_destroy(struct xpanel_ui *ui)
{
	xpw_fini();
	if(ui)
		free(ui);
	xpanel_ui = NULL;
}

static void feedback_lock(struct xpanel_ui *ui)
{
	int backlight = (ui->screen_lock_is_active) ? 0 : 1;

	pcp_backlight(ui->pcpdev, backlight);
}

static void passwd_notify(struct xpw *widget, enum keycodes key, void *data)
{
	struct xpanel_ui	*ui = data;
	struct xpw_passwd	*ui_passwd;
	char			pw_str[100];

	assert(widget->classid = XPW_PASSWD);
	ui_passwd = (struct xpw_passwd *)widget;
	xpw_passwd_read(ui_passwd, pw_str, sizeof(pw_str) - 1);
	DBG("%s: %s:%s\n", __func__, ui->login_name, pw_str);
	if (xpanel_auth_check(ui->login_name, pw_str) == 0) {
		DBG("%s: OK\n", __func__);
		xpw_stack_pop();
		ui->screen_lock_is_active = 0;
		ui->screen_lock_password_active = 0;
		feedback_lock(ui);
		/* finished */
	} else {
		DBG("%s: BAD passwd\n", __func__);
		//xpw_stack_push(WIDGET(ui_passwd));
	}
}

static int password_activate(struct xpanel_ui *ui)
{
	uint64_t	hotkeys_masked;
	int		ret;

	if (ui->screen_lock_password_active)
		return 0;
	if (xpanel_auth_skip(ui->login_name)) {
		INFO("%s: auth_skip\n", __func__);
		return 0;
	}
	INFO("%s: ACTIVATE PASSWORD\n", __func__);
	xpw_hotkey_mask_full(&hotkeys_masked);
	xpw_hotkey_mask_del(&hotkeys_masked, KEYCODE_ESC);
	ret = xpw_passwd_construct(&ui_passwd, ui->login_name, &hotkeys_masked);
	if (ret < 0) {
		ERR("Failed constructing passwd dialog (status %d)\n", ret);
		return ret;
	}
	xpw_passwd_set_facade(&ui_passwd, ui->screen_lock_facade);
	xpw_set_callback(WIDGET(&ui_passwd), passwd_notify, ui);
	xpw_stack_push(WIDGET(&ui_passwd));
	ui->screen_lock_password_active = 1;
	widget_redraw(ui->screen, WIDGET(&ui_passwd));
	return 0;
}

int screen_unlock(struct xpanel_ui *ui)
{
	if (!ui->screen_lock_is_active)
		return 0;
	INFO("%s: de-activate screen lock\n", __func__);
	ui->screen_lock_is_active = 0;
	password_activate(ui);
	feedback_lock(ui);
	return 0;
}

int screen_lock(struct xpanel_ui *ui)
{
	if (ui->screen_lock_is_active)
		return 0;
	INFO("%s: activate screen lock\n", __func__);
	ui->screen_lock_is_active = 1;
	feedback_lock(ui);
	return 0;
}

int xpanel_kbd_idle(struct xpanel_kbd *kbd, void *data)
{
	struct xpanel_ui	*ui = data;
	enum kbd_idle_state	new_state = kbd->idle_state;
	int ret = 0;

	/* INFO("%s (idle_state: %s)\n", __func__, idle_state_name(new_state)); */
	if (ui->screen_lock_password_active && xpanel_auth_skip(ui->login_name)) {
		INFO("%s: BREAK PASSWORD\n", __func__);
		ui->screen_lock_password_active = 0;
		xpw_stack_pop();
		ret = screen_unlock(ui);
		widget_redraw(ui->screen, xpw_current());
		goto out;
	}
	if (new_state == KBD_IDLE_START) {
		/* going to idle */
		ret = screen_lock(ui);
	} else if (new_state == KBD_IDLE_STOP) {
		/* going active */
		ret = screen_unlock(ui);
	}
out:
	return ret;
}
